@@ -4,37 +4,43 @@ use strict;
 use warnings;
 no warnings 'redefine';
 
+my %CHAINED;
+my %ANTICHAINED;
+my %RESTRICT;
+my $UNIV_ISA;
+
 sub generate_CHAINED :Sub(Private)
 {
-    my ($CHAINED,       $ANTICHAINED,
-        $TREE_TOP_DOWN, $TREE_BOTTOM_UP, $u_isa) = @_;
+    my ($chain, $antichain, $TREE_TOP_DOWN, $TREE_BOTTOM_UP, $u_isa) = @_;
+
+    $UNIV_ISA = $u_isa;
 
     # Get names for :CHAINED methods
-    my (%chain, %chain_loc, %chain_restrict);
-    foreach my $package (keys(%{$CHAINED})) {
-        while (my $info = shift(@{$$CHAINED{$package}})) {
+    my (%chain_loc);
+    foreach my $package (keys(%{$chain})) {
+        while (my $info = shift(@{$$chain{$package}})) {
             my ($code, $location, $name, $restrict) = @{$info};
             $name ||= sub_name($code, ':CHAINED', $location);
-            $chain{$name}{$package} = $code;
+            $CHAINED{$name}{$package} = $code;
             $chain_loc{$name}{$package} = $location;
             if ($restrict) {
-                $chain_restrict{$name} = $u_isa;
+                $RESTRICT{$package}{$name} = 1;
             }
         }
     }
 
     # Get names for :CHAINED(BOTTOM UP) methods
     my (%antichain, %antichain_restrict);
-    foreach my $package (keys(%{$ANTICHAINED})) {
-        while (my $info = shift(@{$$ANTICHAINED{$package}})) {
+    foreach my $package (keys(%{$antichain})) {
+        while (my $info = shift(@{$$antichain{$package}})) {
             my ($code, $location, $name, $restrict) = @{$info};
             $name ||= sub_name($code, ':CHAINED(BOTTOM UP)', $location);
 
             # Check for conflicting definitions of $name
-            if ($chain{$name}) {
-                foreach my $other_package (keys(%{$chain{$name}})) {
-                    if ($other_package->$u_isa($package) ||
-                        $package->$u_isa($other_package))
+            if ($CHAINED{$name}) {
+                foreach my $other_package (keys(%{$CHAINED{$name}})) {
+                    if ($other_package->$UNIV_ISA($package) ||
+                        $package->$UNIV_ISA($other_package))
                     {
                         my ($pkg,  $file,  $line)  = @{$chain_loc{$name}{$other_package}};
                         my ($pkg2, $file2, $line2) = @{$location};
@@ -46,9 +52,30 @@ sub generate_CHAINED :Sub(Private)
                 }
             }
 
-            $antichain{$name}{$package} = $code;
+            $ANTICHAINED{$name}{$package} = $code;
             if ($restrict) {
-                $antichain_restrict{$name} = $u_isa;
+                $RESTRICT{$package}{$name} = 1;
+            }
+        }
+    }
+
+    # Propagate restrictions
+    my $reapply = 1;
+    while ($reapply) {
+        $reapply = 0;
+
+        foreach my $pkg (keys(%RESTRICT)) {
+            foreach my $class (keys(%{$TREE_TOP_DOWN})) {
+                if (grep { $_ eq $pkg } @{$$TREE_TOP_DOWN{$class}}) {
+                    foreach my $p (@{$$TREE_TOP_DOWN{$class}}) {
+                        foreach my $n (keys(%{$RESTRICT{$pkg}})) {
+                            if (! exists($RESTRICT{$p}{$n})) {
+                                $RESTRICT{$p}{$n} = 1;
+                                $reapply = 1;
+                            }
+                        }
+                    }
+                }
             }
         }
     }
@@ -57,24 +84,24 @@ sub generate_CHAINED :Sub(Private)
     no strict 'refs';
 
     # Implement :CHAINED methods
-    foreach my $name (keys(%chain)) {
-        my $code = create_CHAINED($TREE_TOP_DOWN, $chain{$name}, $chain_restrict{$name}, $name);
-        foreach my $package (keys(%{$chain{$name}})) {
+    foreach my $name (keys(%CHAINED)) {
+        my $code = create_CHAINED($name, $TREE_TOP_DOWN, $CHAINED{$name});
+        foreach my $package (keys(%{$CHAINED{$name}})) {
             *{$package.'::'.$name} = $code;
             add_meta($package, $name, 'kind', 'chained');
-            if ($chain_restrict{$name}) {
+            if ($RESTRICT{$package}{$name}) {
                 add_meta($package, $name, 'restricted', 1);
             }
         }
     }
 
     # Implement :CHAINED(BOTTOM UP) methods
-    foreach my $name (keys(%antichain)) {
-        my $code = create_CHAINED($TREE_BOTTOM_UP, $antichain{$name}, $antichain_restrict{$name}, $name);
-        foreach my $package (keys(%{$antichain{$name}})) {
+    foreach my $name (keys(%ANTICHAINED)) {
+        my $code = create_CHAINED($name, $TREE_BOTTOM_UP, $ANTICHAINED{$name});
+        foreach my $package (keys(%{$ANTICHAINED{$name}})) {
             *{$package.'::'.$name} = $code;
             add_meta($package, $name, 'kind', 'chained (bottom up)');
-            if ($antichain_restrict{$name}) {
+            if ($RESTRICT{$package}{$name}) {
                 add_meta($package, $name, 'restricted', 1);
             }
         }
@@ -86,11 +113,10 @@ sub generate_CHAINED :Sub(Private)
 # and CHAINED(BOTTOM UP) methods for a particular method name.
 sub create_CHAINED :Sub(Private)
 {
+    # $name      - method name
     # $tree      - ref to either %TREE_TOP_DOWN or %TREE_BOTTOM_UP
     # $code_refs - hash ref by package of code refs for a particular method name
-    # $restrict  - restricted method (trick: == $UNIV_ISA)
-    # $name      - method name
-    my ($tree, $code_refs, $restrict, $name) = @_;
+    my ($name, $tree, $code_refs) = @_;
 
     return sub {
         my $thing = shift;
@@ -100,9 +126,9 @@ sub create_CHAINED :Sub(Private)
         my @classes;
 
         # Caller must be in class hierarchy
-        if ($restrict) {
+        if ($RESTRICT{$class}{$name}) {
             my $caller = caller();
-            if (! ($caller->$restrict($class) || $class->$restrict($caller))) {
+            if (! ($caller->$UNIV_ISA($class) || $class->$UNIV_ISA($caller))) {
                 OIO::Method->die('message' => "Can't call restricted method '$class->$name' from class '$caller'");
             }
         }
@@ -125,5 +151,5 @@ sub create_CHAINED :Sub(Private)
 
 
 # Ensure correct versioning
-my $VERSION = 2.22;
-($Object::InsideOut::VERSION == 2.22) or die("Version mismatch\n");
+my $VERSION = 2.23;
+($Object::InsideOut::VERSION == 2.23) or die("Version mismatch\n");
